# Realtime raytracing using POV-Ray and Syzygy
# This C++ code writes a .pov scene file,
# calls povray to turn it into a .ppm image file,
# loads the ppm into a texture,
# displays the texture on a quadrilateral on the wall.
#
# Written by djzielin@duke.edu 5-2006
# Tweaked by Camille Goudeseune.
# Python version by Jim Crowell

from PySZG import *
from OpenGL.GL import *
from OpenGL.GLU import *
import os
import time
import math

__coordTransMatrix = arMatrix4( (1,0,0,0,0,1,0,0,0,0,-1,0,0,0,0,1) )

# Convert a matrix from syzygy to povray coordinates
def ar_szgMatrixToPOV( szgMatrix ):
   return __coordTransMatrix * szgMatrix * __coordTransMatrix

# Convert an arMatrix4 in povray coords to a povray matrix tuple
# (i.e. suitable for a python template of the povray 'matrix'
# command).
def ar_povMatrixToTuple( povMatrix ):
   return tuple([povMatrix[i] for i in [0,1,2,4,5,6,8,9,10,12,13,14]])

class POVScene(object):
   def __init__(self):
     object.__init__(self)
     # (period, amplitude, offset) tuples
     self.lissData = [(.7,5.,-2.),(.3,5.,12.)]
     self.phase = 0.
     self.wandMatrixTuple = (1,0,0,0,1,0,0,0,1,1,4,1)
   def update( self, app ):
     self.wandMatrixTuple = ar_povMatrixToTuple( app.getMatrix(1) )
     self.phase += .1
   def getState(self):
     return (self.phase,)+self.wandMatrixTuple
   def setState( self, stateTuple ):
     self.phase = stateTuple[0]
     self.wandMatrixTuple = stateTuple[1:]
   def __call__(self):
     self.liss = [i[2]+i[1]*math.sin(self.phase/i[0]) for i in self.lissData]
     # See www.povray.org/documentation/ .
     return [\
"""#include \"colors.inc\"
#include \"shapes.inc\"
#include \"textures.inc\"
#include \"glass.inc\"
#include \"consts.inc\"
light_source { <2,8,-2> color White }
light_source { <%f,%f,-2> color White }
light_source { <-4,13,-3> color Red }
background { color Blue }
sphere { <0,5,3>, .5 pigment {Magenta} finish {ambient .2 diffuse .5
specular .5 }}
""" % tuple(self.liss), \
"""
box {<-.1,-.1,0>,<.1,.1,5> pigment {White} matrix <%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f>}
plane { y, 0 pigment {White} finish { ambient 0.2 diffuse 0.3 } }
""" % self.wandMatrixTuple ]

class POVRenderer(object):
   def __init__(self, scene=POVScene(), textureWidth=128, \
       tmpDirPath='C:/szg', \
       povrayPath='C:/aszgard/binextra/megapov/bin/megapov.exe', \
       povLibPath='C:/aszgard/binextra/megapov/include' ):
     if not textureWidth in [2**i for i in range(1,15)]:
       raise ValueError, 'POVRenderer(): textureWidth must be a power of 2 (> 1).'
     object.__init__(self)
     self.sceneGenerator = scene
     self.textureWidth = textureWidth
     self.povrayPath = povrayPath
     self.povLibPath = povLibPath
     self.tmpDirPath = tmpDirPath
     self.povFilename = os.path.join( tmpDirPath, 'szgpov.pov' )
     self.ppmFilename = os.path.join( tmpDirPath, 'szgpov.ppm' )
     self.texture  = arTexture()
     print 'szgpov POV-ray path:',self.povrayPath
     print 'szgpov POV-ray library directory:',self.povLibPath
     print 'szgpov Output file:',self.povFilename
     print 'szgpov Input file:',self.ppmFilename
     self.secondsPrev = 0
     self.clockSetting = 0.0

   def __del__(self):
     try:
       print 'szgpov removing',self.povFilename
       # os.remove( self.povFilename )
     except Exception:
       pass
     try:
       print 'szgpov removing',self.ppmFilename
       # os.remove( self.ppmFilename )
     except Exception:
       pass

   def setViewParameters( self, eyePosition, \
         centerScreen, normalScreen, upScreen, \
         widthScreen, heightScreen, \
         widthWindow, heightWindow ):
     # extract some view-related quantities that are used many places
     self.eyePosition = eyePosition
     self.centerScreen = centerScreen
     self.normalScreen = normalScreen
     self.upScreen = upScreen
     # * = cross product. Left-handed coordinates.
     self.rightScreen = self.upScreen * self.normalScreen
     self.widthScreen = widthScreen
     self.heightScreen = heightScreen
     self.widthWindow = widthWindow
     self.heightWindow = heightWindow

   def setTextureWidth(self, w):
     self.textureWidth = w

   def setPOVClock(self, clockVal):
     self.clockSetting = clockVal

   def draw(self):
     # Generate the .pov file
     self._writeFile()
     # call povray, read its generated .ppm file
     if not self._povRender():
       raise RuntimeError, 'szgpov _povRender() failed.'
     # Override Syzygy's view matrix.
     self._loadProjection()
     # draw the image
     self._drawTexture()
     self._unloadProjection()
     self._drawMetrics()

   def __str__(self):
     # the sceneGenerator _must_ return a list of strings
     return '\n'.join( self.sceneGenerator() + self._cameraStringList() )

   def _writeFile(self):
     outFile = file( self.povFilename, "w" )
     outFile.write( str(self) )
     outFile.close()

   def _cameraStringList(self):
     # The camera direction, up, and right parameters set the field of view.
     # According to the povray docs, they obey:
     # tan(xFOV/2) = right.magnitude()/(2*direction.magnitude())
     # tan(yFOV/2) = up.magnitude()/(2*direction.magnitude())
     upCamera = self.upScreen * self.heightScreen
     rightCamera = self.rightScreen * self.widthScreen
     # project eye-screen vector onto screen normal
     directionCamera = self.centerScreen - self.eyePosition
     return ['camera {\nlocation <%f,%f,%f>'  % self.eyePosition.toTuple(), \
         'direction <%f,%f,%f>' % directionCamera.toTuple(), \
         'up <%f,%f,%f>'        % upCamera.toTuple(), \
         'right <%f,%f,%f>\n /*normal{facets size 0.06 }*/}'   % rightCamera.toTuple()]

   # Call povray and read its output.
   def _povRender(self):
     commandString = \
       r"%s -D -H%d -W%d -I%s +FP +O%s +L%s -A +GD +GAc:\szg\povstreams.txt -UV +K%f" % \
       (self.povrayPath, \
           self.textureWidth, self.textureWidth, \
           self.povFilename, self.ppmFilename, self.povLibPath,\
           self.clockSetting)
     (inFile,outFile,errFile) = os.popen3( commandString )
     errTxt = errFile.read()
     lineList = outFile.readlines(200) # 200 is max # lines to read
     inFile.close()
     outFile.close()
     errFile.close()
     if not self.texture.readPPM( self.ppmFilename, '', '' ):
       print errTxt
       print 'szgpov failed to read texture file.'
       return False
#   # Parse megapov's output.  Get the (only) line starting with 'Bounding Box'.
#   bboxLine = [l for l in lineList if l.find('Bounding Box')==0][0]
#   # complains "list index out of range" --^
#   bboxTokens = [w for w in bboxLine.split(' ') if w]
#   self.numtestsBbox = int(bboxTokens[2])
     self.numtestsBbox = 424242
     return True

   def _makePPMFileName(self):
     val = 'szgpov_'+str(self.frameNum)+'.ppm'
     self.frameNum += 1
     if self.frameNum > 99:
       self.frameNum = 0
     return val

   def _loadProjection(self):
     self.projectionMatrix = glGetFloat(GL_PROJECTION_MATRIX)
     self.modelMatrix = glGetFloat(GL_MODELVIEW_MATRIX)
     glMatrixMode(GL_PROJECTION)
     glLoadIdentity()
     gluOrtho2D(0., 1., 0., 1.)  # 2d mode
     glViewport(0, 0, self.widthWindow, self.heightWindow)
     glMatrixMode(GL_MODELVIEW)
     glLoadIdentity()

   def _unloadProjection(self):
     glMatrixMode(GL_PROJECTION)
     glLoadMatrixf(self.projectionMatrix)
     glMatrixMode(GL_MODELVIEW)
     glLoadMatrixf(self.modelMatrix)

   def _drawTexture(self):
     self.texture.activate()
     glBegin(GL_QUADS)
     glTexCoord2f(1.,0.)
     glVertex2f(0.,0.)
     glTexCoord2f(0.,0.)
     glVertex2f(1.,0.)
     glTexCoord2f(0.,1.)
     glVertex2f(1.,1.)
     glTexCoord2f(1.,1.)
     glVertex2f(0.,1.)
     glEnd()
     self.texture.deactivate()

   def _glutprintfinit(self):
     # wide antialiased lines
     #glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
     #glEnable(GL_BLEND)
     #glEnable(GL_LINE_SMOOTH)
     glDisable(GL_DEPTH_TEST)
     glLineWidth(3.0)
     glColor3f( 1.0, 1.0, 0.4 )

   def _glutprintfend(self):
     glEnable(GL_DEPTH_TEST)
     glLineWidth(1)

   def _glutprintf(self, x, y, z, conststring, rotY, rotX):
     from OpenGL.GLUT import glutStrokeCharacter, GLUT_STROKE_MONO_ROMAN, GLUT_STROKE_ROMAN
     glPushMatrix()
     glTranslatef(x, y, z)
     glScalef(.0022,.0022,.0022)
     glRotatef(rotY, 0, -1, 0)
     glRotatef(rotX, 1,  0, 0)
     for c in conststring:
       glutStrokeCharacter(GLUT_STROKE_ROMAN, ord(c))
     glPopMatrix()

   def _drawMetrics(self):
     # Compute frames per second.
     secondsNow = time.clock()
     spf = secondsNow - self.secondsPrev
     self.secondsPrev = secondsNow
     if spf < 0.001:
       # Avoid division by zero.
       spf = 0.001
     fps = str(1.0 / spf)
     # todo: draw left and right frame fps not on top of each other
#   tmp = "bbox tests = " + str(self.numtestsBbox) + ", Camille hopes"
     self._glutprintfinit()
#   self._glutprintf(0,  6, -2.9, fps + "fps" , 0, 0)
#   self._glutprintf(0,  6, -3.9, tmp                 , 0, 0)
     self._glutprintf(0,  7, -3.9, "Front wall, " + fps + " fps", 0, 0)
     self._glutprintf(-4.9, 5,  0, "Left wall, " + fps + " fps",  -90, 0)
     self._glutprintf( 4.9, 5,  0, "Right wall, " + fps + " fps",  90, 0)
     self._glutprintf(0,  5,  4.9, "Back wall, " + fps + " fps",  180, 0)
     self._glutprintf(0,  9.9,  0, "Top wall, " + fps + " fps", 0, 90)
     self._glutprintf(0,  0.1,  0, "Bottom wall, " + fps + " fps", 0, -90)
     self._glutprintfend()

class POVApp(arPyMasterSlaveFramework):
   def __init__( self, scene=POVScene(), renderer=None ):
     arPyMasterSlaveFramework.__init__(self)
     from optparse import OptionParser
     parser = OptionParser()
     parser.add_option( '-d', '--tmpdirpath', \
           action='store', type='string', dest='tmpDirPath', \
           help='Path to directory in which to create temporary files'+ \
                '(default=c:/szg; use forward slashes).' )
     parser.add_option( '-p', '--povraypath', \
           action='store', type='string', dest='povrayPath', \
           help='Path to megapov executable (including .exe on Windows)'+ \
                '(default=c:/povray_3.6/bin/megapov.exe; use forward slashes).' )
     parser.add_option( '-l', '--povlibpath', \
           action='store', type='string', dest='povLibPath', \
           help='Path to povray include directory '+ \
                '(default=c:/povray_3.6/include; use forward slashes).')
     parser.add_option( '-w', '--texturewidth', \
           action='store', type='int', dest='textureWidth', \
           help='Width and height of rendered image in pixels (default=128).' )
     (options,args) = parser.parse_args() # parses sys.argv[1:] by default
     optDict = {}
     for k in ['tmpDirPath','povrayPath','povLibPath','textureWidth']:
       val = getattr( options, k )
       if val:
         optDict[k] = val
     optDict['scene'] = scene
     self.scene = scene
     if renderer:
       self.renderer = renderer
     else:
       self.renderer = POVRenderer( **optDict )
     self.textureWidth = 128
     self.clockSetting = 0
     self.clockDelta = 0.01
   def onStart( self, client ):
     self.setClipPlanes(0.1, 100)
     self.setUnitConversion(1.0)
     self.initSequenceTransfer('scene')
     self.initSequenceTransfer('clSwitches')
     self.eyePrev = arVector3(0,0,0)
     return True
   def getMatrix( self, index ):
     return ar_szgMatrixToPOV( arPyMasterSlaveFramework.getMatrix( self, index ) )
   def onPreExchange(self):
     self.scene.update( self )
     self.setSequence( 'scene', self.scene.getState() )
     if self.getOnButton(0):
       # do something with left button
       foo = 0
     if self.getOnButton(1):
       # do something with middle button
       foo = 0
     if self.getOnButton(2):
       # Right button cycles through screen resolutions.
       self.textureWidth = self.textureWidth / 2
       if self.textureWidth < 32:
         self.textureWidth = 256
     self.clockSetting += self.clockDelta
     if (self.clockSetting >1.0):
       self.clockSetting = 0
     switchSequence = (self.textureWidth, self.clockSetting)
     self.setSequence( 'clSwitches', switchSequence )
     self.renderer.setTextureWidth(self.textureWidth)
     self.renderer.setPOVClock(self.clockSetting)

   def onPostExchange(self):
     if not self.getMaster():
       self.scene.setState( self.getSequence('scene') )
       clSequence = self.getSequence( 'clSwitches' )
       textureWidth = clSequence[0]
       clockVal = clSequence[1]
       self.renderer.setTextureWidth(textureWidth)
       self.renderer.setClock(clockVal)
     
   def onDraw( self, win, vp ):
     (x,y,widthWindow,heightWindow) = win.getPixelDimensions()
     screen = vp.getScreen()
     center = screen.getCenter()
     normal = screen.getNormal()
     headMounted = screen.getHeadMounted()
     up = screen.getUp()
     if headMounted:
       midEyeMatrix = self.getMidEyeMatrix()
       center = midEyeMatrix * center
       orient = ar_extractRotationMatrix( midEyeMatrix )
       normal = orient * normal;
       up = orient * up;
     eyePos = self.getHead().getEyePosition( vp.getEyeSign() )
     center[2] = -center[2]
     normal[2] = -normal[2]
     up[2] = -up[2]
     eyePos[2] = -eyePos[2]
     #Higher resolution when head moves slower (for static scenes).
     # todo: this is feet per frame, should be feet per second
     # todo: in "szgpov.py" app, only left eye of north screen
     # changes.  Should be entire cube all together.  Surely self.getHead() is
     # consistent across all screens?
     eyeSpeed = (eyePos - self.eyePrev).magnitude()
     self.eyePrev = eyePos
  #  if eyeSpeed > 0.25:
  #    self.renderer.setTextureWidth( 16 )
  #  else:
  #    self.renderer.setTextureWidth( 128)
     self.renderer.setViewParameters( eyePos, \
         center, normal, up, \
         screen.getWidth(), screen.getHeight(), \
         widthWindow, heightWindow )
     self.renderer.draw()
#    print self.renderer
   def onExit(self):
     print 'szgpov exiting.'
     self.renderer = None
     self.scene = None

if __name__=='__main__':
   import sys
   app = POVApp()
   if not app.init( sys.argv ):
     raise RuntimeError, 'szgpov init failed.'
   if not app.start():
     raise RuntimeError, 'szgpov start failed.'
